unit Main;

{
    Author  : Silicoid
    Date    : 14.07.2005
    Version : 1.1
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, StrUtils, ExtCtrls, ImgList;

const
  FLAG_TREE_DEF = $FF;
  FLAG_TREE_NO  = 0;
  FLAG_TREE_YES = 1;

  FLAG_DATA_TEXT = 1;
  FLAG_DATA_TREE = 2;

  IMAGE_DEFAULT = 0;
  IMAGE_SELECT  = 1;

type
  TItem = class(TObject)
  private
    isTree  : Byte;
    count   : Integer;
    //---------------------------
    unknow  : Integer;
    flag    : Integer;
    isText  : Byte;
    NickName: WideString;
    DataText: WideString;
    mas     : array of TItem;
    //---------------------------
    nodeLink: TTreeNode;
    isSelect: Boolean;
  public
    constructor Create();
    destructor Free();
    procedure Clear();
    //----------------------------
    function  AddNewItem(): TItem;
    procedure ReadTree(pFile: TFileStream);
    procedure WriteTree(pFile: TFileStream);
    procedure ReadText(pFile: TFileStream);
    procedure WriteText(pFile: TFileStream);
    procedure ReadData(pFile: TFileStream);
    procedure WriteData(pFile: TFileStream);
    procedure AddItemToNode(trvMas: TTreeView; node: TTreeNode);
    function  GetAllChildText(): WideString;
    function  GetSize(): Integer;
    function  GetImage(): Integer;
    procedure SetImage();
    function  FindStr(s: WideString; var cntRes: Cardinal): Boolean;
    procedure SetSelect(isSel: Boolean);
  end;

  TFrmMain = class(TForm)
    dlgOpen : TOpenDialog;
    prgWork : TProgressBar;
    stsWork : TStatusBar;
    pnlData : TPanel;
    grpCmd  : TGroupBox;
    btnLoad : TButton;
    btnSave : TButton;
    grpText : TGroupBox;
    memText : TMemo;
    pnlTree : TPanel;
    grpTree : TGroupBox;
    trvMas  : TTreeView;
    splMain : TSplitter;
    grpDat  : TGroupBox;
    lblDat  : TLabel;
    edtDat  : TEdit;
    btnDat  : TButton;
    chkNoSum: TCheckBox;
    dlgSave : TSaveDialog;
    edtFind : TEdit;
    btnFind : TButton;
    btnReset: TButton;
    imlTree : TImageList;
    procedure btnDatClick(Sender: TObject);
    procedure btnLoadClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure trvMasClick(Sender: TObject);
    procedure btnSaveClick(Sender: TObject);
    procedure memTextChange(Sender: TObject);
    procedure trvMasKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure btnFindClick(Sender: TObject);
    procedure edtFindKeyPress(Sender: TObject; var Key: Char);
    procedure btnResetClick(Sender: TObject);
    procedure trvMasGetSelectedIndex(Sender: TObject; Node: TTreeNode);
  public
    startItem: TItem;
    nameFile : String;
    //------------------------------------
    procedure EnableMain(isEn: Boolean);
    function  StartRead(sFile: String): Boolean;
    procedure StartWrite(sFile: String);
    procedure BuildTree();
    procedure OnClickTree();
    procedure SetTextInfo(s: String);
    procedure SetIsEditText(isEd: Boolean);
    procedure OnMemoChange();
  end;

var
  FrmMain: TFrmMain;

implementation

{$R *.dfm}

function GetWideZero(pFile: TFileStream): WideString;
var
  s: WideString;
  c: WideChar;
begin
  s:='';
  repeat
    pFile.Read(c,2);
    if c<>#0 then s:=s+c;
  until c=#0;
  Result:=s;
end;

procedure PutWideZero(pFile: TFileStream; s: WideString);
var
  i: Integer;
  c: WideChar;
begin
  for i:=1 to Length(s) do
  begin
    c:=s[i]; pFile.Write(c,2); 
  end;
  c:=#0;
  pFile.Write(c,2);
end;

procedure Msg(s: String); overload;
begin
  Application.MessageBox(PChar(s),'Debug',MB_OK); 
end;

procedure Msg(k: Integer); overload;
begin
  Application.MessageBox(PChar(IntToStr(k)),'Debug',MB_OK); 
end;

procedure InitProgress(mx: Integer; s:String);
begin
  with FrmMain do
  begin
    prgWork.min:=0;
    prgWork.Max:=mx;
    prgWork.Position:=0;
    stsWork.SimpleText:=s;
  end;
end;

procedure SetProgress(cur: Integer);
begin
  with FrmMain do
  begin
    prgWork.Position:=cur;
  end;
  Application.ProcessMessages;
end;

procedure IncProgress();
var
  cur: Integer;
begin
  with FrmMain.prgWork do
  begin
    cur:=Position+1;
    if cur<Max then Position:=cur;
  end;
  Application.ProcessMessages;  
end;

procedure SetStatus(s: String);
begin
  FrmMain.stsWork.SimpleText:=s;
  Application.ProcessMessages;
end;

function ExtractWithoutExt(sPath: string):string;
var
  nam, ext:string;
begin
  nam:=ExtractFileName(sPath);
  ext:=ExtractFileExt(sPath);
  Result:=LeftStr(nam, Length(nam)-Length(ext));
end;

//--------------------------------------------

constructor TItem.Create();
begin
  Clear();
end;

destructor TItem.Free;
begin
  Clear();
end;

procedure TItem.Clear();
var
  i: Integer;
begin
  isTree  :=FLAG_TREE_DEF;
  count   :=0;
  //----------------------
  unknow  :=0;
  flag    :=1;
  isText  :=0;
  NickName:='';
  DataText:='';
  isSelect:=false;
  //----------------------
  for i:=0 to High(mas) do if mas[i]<>nil then mas[i].Free();
  SetLength(mas,0);
end;

function TItem.AddNewItem(): TItem;
var
  ln: Integer;
begin
  ln:=Length(mas);
  SetLength(mas,ln+1);
  mas[ln]:=TItem.Create;
  Result:=mas[ln];
end;

procedure TItem.ReadTree(pFile: TFileStream);
var
  item: TItem;
  i: Integer;
begin
  pFile.Read(isTree,1);
  if isTree=FLAG_TREE_YES then
  begin
    pFile.Read(count,4);
    for i:=1 to count do
    begin
      item:=AddNewItem();
      item.ReadText(pFile);
    end;
  end else
  begin
    pFile.Read(count,4);
    for i:=1 to count do
    begin
      item:=AddNewItem();
      item.ReadData(pFile);
    end;
  end;
  //------------------------
  SetProgress(pFile.Position);
end;

procedure TItem.WriteTree(pFile: TFileStream);
var
  item: TItem;
  i: Integer;
begin
  pFile.Write(isTree,1);
  if isTree=FLAG_TREE_YES then
  begin
    pFile.Write(count,4);
    for i:=0 to High(mas) do
    begin
      item:=mas[i];
      item.WriteText(pFile);
    end;
  end else
  begin
    pFile.Write(count,4);
    for i:=0 to High(mas) do
    begin
      item:=mas[i];
      item.WriteData(pFile);
    end;
  end;
  //------------------------
  IncProgress();
end;

procedure TItem.ReadText(pFile: TFileStream);
begin
  pFile.Read(unknow,4);
  pFile.Read(flag,4);
  pFile.Read(isText,1);
  NickName:=GetWideZero(pFile);
  if isText=FLAG_DATA_TREE then
  begin
    ReadTree(pFile);
  end else
  begin
    DataText:=GetWideZero(pFile);
  end;
  //------------------------
  SetProgress(pFile.Position);
end;

procedure TItem.WriteText(pFile: TFileStream);
begin
  pFile.Write(unknow,4);
  pFile.Write(flag,4);
  pFile.Write(isText,1);
  PutWideZero(pFile,NickName);
  if isText=FLAG_DATA_TREE then
  begin
    WriteTree(pFile);
  end else
  begin
    PutWideZero(pFile,DataText);
  end;
  //------------------------
  IncProgress();
end;

procedure TItem.ReadData(pFile: TFileStream);
begin
  pFile.Read(isText,1);
  NickName:=GetWideZero(pFile);
  DataText:=GetWideZero(pFile);
  //------------------------
  SetProgress(pFile.Position);
end;

procedure TItem.WriteData(pFile: TFileStream);
begin
  pFile.Write(isText,1);
  PutWideZero(pFile,NickName);
  PutWideZero(pFile,DataText);
  //------------------------
  IncProgress();
end;

procedure TItem.AddItemToNode(trvMas: TTreeView; node: TTreeNode);
var
  i: Integer;
begin
  nodeLink:=trvMas.Items.AddChildObject(node, NickName,self);
  //----------------------
  for i:=0 to High(mas) do
  begin
    mas[i].AddItemToNode(trvMas, nodeLink); 
  end;
  //------------------------
  IncProgress();
end;

function TItem.GetAllChildText(): WideString;
var
  s,t: WideString;
  i: Integer;
begin
  s:='';
  for i:=0 to High(mas) do
  begin
    t:=mas[i].DataText;
    if Length(t)<>0 then s:=s+t+#13#10+#13#10;
    s:=s+mas[i].GetAllChildText();
  end;
  Result:=s;  
  //------------------------
  IncProgress();
end;

function TItem.GetSize(): Integer;
var
  z,i: Integer;
begin
  z:=1;
  for i:=0 to High(mas) do
  begin
    z:=z+mas[i].GetSize();
  end;
  Result:=z;
end;

function TItem.GetImage(): Integer;
begin
  if isSelect then Result:=IMAGE_SELECT else Result:=IMAGE_DEFAULT;
end;

procedure TItem.SetImage();
begin
  nodeLink.ImageIndex:=GetImage();
end;

function TItem.FindStr(s: WideString; var cntRes: Cardinal): Boolean;
var
  i: Integer;
begin
  isSelect:=false;
  for i:=0 to High(mas) do
  begin
    if mas[i].FindStr(s,cntRes) then isSelect:=true; 
  end;
  if Pos(s, WideUpperCase(DataText))<>0 then
  begin
    isSelect:=true; cntRes:=cntRes+1;
  end;
  Result:=isSelect; SetImage(); IncProgress();
end;

procedure TItem.SetSelect(isSel: Boolean);
var
  i: Integer;
begin
  isSelect:=isSel;
  for i:=0 to High(mas) do
  begin
    mas[i].SetSelect(isSel);  
  end;
  SetImage(); IncProgress();
end;

//------------------------------------------

function TFrmMain.StartRead(sFile: String): Boolean;
var
  ret: Integer;
  pFile: TFileStream;
  ch: Cardinal;
begin
  ret:=IDYES; Result:=false;
  if startItem<>nil then
  begin
    ret:=Application.MessageBox('    .  ? ', '', MB_YESNO);
  end;
  if ret=IDYES then
  begin
    if FileExists(sFile) then
    begin
      memText.Clear; trvMas.Items.Clear;
      pFile:=TFileStream.Create(sFile,fmOpenRead);
      try
        pFile.Seek(0,soFromBeginning);
        pFile.Read(ch,4);
        if ch<>$31304C5A then
        begin
          nameFile:=ExtractFileName(edtDat.Text);
          InitProgress(pFile.Size,'  ,  ...');
          pFile.Seek(0,soFromBeginning);
          if startItem<>nil then startItem.Free;
          startItem:=TItem.Create;
          startItem.ReadTree(pFile);
          InitProgress(0,'');
          Result:=true;
        end else
        begin
          Application.MessageBox('  !   .','',MB_OK);
        end;
      finally
        pFile.Free;
      end;
    end else
    begin
      Application.MessageBox('  ,    .','',MB_OK);
    end;
  end;
end;

procedure TFrmMain.StartWrite(sFile: String);
var
  pFile: TFileStream;
begin
  pFile:=TFileStream.Create(sFile, fmCreate);
  try
    InitProgress(startItem.GetSize(),'  , ...');
    startItem.WriteTree(pFile);
    InitProgress(0,'');
  finally
    pFile.Free;
  end;
end;

procedure TFrmMain.BuildTree();
var
  topNode: TTreeNode;
begin
  if startItem<>nil then
  begin
    InitProgress(startItem.GetSize(),'  ,  ...');
    trvMas.Items.Clear;
    startItem.isText:=FLAG_DATA_TREE;
    startItem.AddItemToNode(trvMas, nil);
    topNode:=trvMas.TopItem;
    if Length(topNode.Text)=0 then topNode.Text:=nameFile;
    topNode.Expand(false);
    InitProgress(0,'');
  end;
end;

procedure TFrmMain.SetTextInfo(s: String);
begin
  FrmMain.grpText.Caption:=s;
end;

procedure TFrmMain.SetIsEditText(isEd: Boolean);
begin
  FrmMain.memText.ReadOnly:=not isEd;
end;

procedure TFrmMain.OnClickTree();
var
  item: TItem;
  selNode: TTreeNode;
  s: WideString;
begin
  selNode:=trvMas.Selected;
  if selNode<>nil then
  begin
    memText.Clear;
    if selNode.Data<>nil then
    begin
      item:=TItem(selNode.Data);
      s:=item.DataText;
      //--------------------------------
      if item.isText=FLAG_DATA_TREE then
      begin
        if not chkNoSum.Checked then
        begin
          InitProgress(item.GetSize(),'  , ...');
          s:=item.GetAllChildText();
          SetTextInfo(' (  )');
          SetIsEditText(false);
          InitProgress(0, '');
        end;
      end else
      begin
        SetIsEditText(true);
        SetTextInfo('');
      end;
      memText.Lines.Text:=s;
    end;
  end;
end;

procedure TFrmMain.OnMemoChange();
var
  item: TItem;
  selNode: TTreeNode;
begin
  selNode:=trvMas.Selected;
  if selNode<>nil then
  begin
    if selNode.Data<>nil then
    begin
      item:=TItem(selNode.Data);
      item.DataText:=memText.Lines.Text;
    end;    
  end;
end;

procedure TFrmMain.EnableMain(isEn: Boolean);
begin
  btnLoad.Enabled :=isEn;
  btnSave.Enabled :=isEn;
  memText.Enabled :=isEn;
  btnFind.Enabled :=isEn;
  btnReset.Enabled:=isEn;
  edtFind.Enabled :=isEn;
end;

//-----------------------------------

procedure TFrmMain.btnDatClick(Sender: TObject);
begin
  if dlgOpen.Execute then edtDat.Text:=dlgOpen.FileName;
end;

procedure TFrmMain.btnLoadClick(Sender: TObject);
begin
  EnableMain(false);
  if StartRead(edtDat.Text) then
  begin
    BuildTree();
    EnableMain(true);
    Application.MessageBox(' ','',MB_OK);
  end;
  EnableMain(true);
end;

procedure TFrmMain.FormCreate(Sender: TObject);
begin
  startItem:=nil;
  nameFile:='';
end;

procedure TFrmMain.trvMasClick(Sender: TObject);
begin
  OnClickTree();
end;

procedure TFrmMain.btnSaveClick(Sender: TObject);
begin
  if dlgSave.Execute then
  begin
    EnableMain(false);
    StartWrite(dlgSave.FileName);
    EnableMain(true);
    Application.MessageBox(PChar(''+#13#10+dlgSave.FileName),'',MB_OK);
  end;
end;

procedure TFrmMain.memTextChange(Sender: TObject);
begin
  if not memText.ReadOnly then OnMemoChange();
end;

procedure TFrmMain.trvMasKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  OnClickTree();
end;

procedure TFrmMain.btnFindClick(Sender: TObject);
var
  s: String;
  cntRes: Cardinal;
begin
  EnableMain(false);
  if startItem<>nil then
  begin
    s:=WideUpperCase(edtFind.Text);
    InitProgress(startItem.GetSize(),'  ,  ...');
    startItem.FindStr(s,cntRes);
    trvMas.Repaint;
    InitProgress(0,'');
    if cntRes=0
    then stsWork.SimpleText:='  '
    else stsWork.SimpleText:='  : ' + IntToStr(cntRes);
  end;
  EnableMain(true);
end;

procedure TFrmMain.btnResetClick(Sender: TObject);
begin
  if startItem<>nil then
  begin
    InitProgress(startItem.GetSize(),'   ,  ...');
    startItem.SetSelect(false);
    trvMas.Repaint;
    InitProgress(0,'');
  end;
end;

procedure TFrmMain.edtFindKeyPress(Sender: TObject; var Key: Char);
begin
  if Key=#13 then btnFind.Click;
end;

procedure TFrmMain.trvMasGetSelectedIndex(Sender: TObject; Node: TTreeNode);
begin
  Node.SelectedIndex:=Node.ImageIndex;
end;

end.


